import os
from copy import deepcopy

from piperabm.model import Model
from piperabm.model.measurement.accessibility import Accessibility
from piperabm.model.measurement.travel_distance import TravelDistance
#from piperabm.model.measurement.interaction import Interaction
from piperabm.tools.json_file import JsonFile
from piperabm.tools.coordinate import distance as ds


class Measurement:

    type = "measurement"

    def __init__(self, path=None, name: str = ''):
        self.path = path
        self.name = name
        self.times = []
        self.accessibility = Accessibility(measurement=self)
        self.travel_distance = TravelDistance(measurement=self)
        #self.interaction = Interaction(measurement=self)
        
    @property
    def result_directory(self):
        """
        Return result directory
        """
        if self.path is None:
            raise ValueError("define path to continue")
        result = os.path.join(self.path, 'result')
        if self.name != '':
            result = os.path.join(result, self.name)
        return result

    @property
    def len(self) -> int:
        """
        Return total number of entries
        """
        return len(self.times) - 1
    
    def add_time(self, value: float) -> None:
        """
        Add new point in time
        """
        self.times.append(deepcopy(value))

    def filter_times(self, _from: int = None, _to: int = None) -> list:
        """
        Filter times list
        """
        times = self.times[1:]
        if _from is None:
            _from = 0
        if _to is None:
            _to = len(times)
        return times[_from:_to]
    
    def delta_times(self, _from: int = None, _to: int = None) -> list:
        """
        Return delta times
        """
        times = self.times[1:]
        if _from is None:
            _from = 0
        if _to is None:
            _to = len(times)
        result = []
        for i in range(_from, _to):
            delta = self.times[i + 1] - self.times[i]
            result.append(delta)
        return result

    def add_accessibility(self, id: int, value: float) -> None:
        """
        Add new accessibility value
        """
        self.accessibility.add(id=id, value=value)
    
    def add_travel_distance(self, value: float) -> None:
        """
        Add new travel distance value
        """
        self.travel_distance.add(value=value)
    '''
    def add_interaction(self, id_from: int, id_to: int, resource_name: str, resource_amount: float) -> None:
        """
        Add new interaction value
        """
        self.interaction.add(id_from=id_from, id_to=id_to, resource_name=resource_name, resource_amount=resource_amount)
    '''
    def measure(self, report=True, resume=False):
        if resume is False:  # Restart the measurement
            file = JsonFile(path=self.path, filename='measurement')
            #if file.exists() is True:
            file.remove()  # Delete the previous measurement if exists
            model = Model(path=self.path, name=self.name)
            model.load_initial()  # Load initial state of model
            self.add_time(model.time)  # First date entry
        elif resume is True:  # Continue the measurement
            raise ValueError("Feature not tested yet.")
            '''
            file = JsonFile(path=self.path, filename=self.name+'_'+'measurement')
            if file.exists() is False:  # Restart the measurement if previous doesn't exist
                self.measure(report=report, resume=False)
            else:
                model = Model(path=self.path, name=self.name)
                model.load_final()  # Load final state of model
                self.load()  # Load final state of measurement
            '''

        deltas = model.load_deltas()
        previous = deepcopy(model)
        n = len(deltas)
        for i, delta in enumerate(deltas):
            model.apply_delta(delta) # Push model a single step forward
            self.add_time(model.time)
            total_travel = 0
            for id in model.society.agents:
                # Accessibility
                self.add_accessibility(
                    id=id,
                    value=model.society.accessibility(id=id)
                )
                # Travel distance
                previous_pos = previous.society.get_pos(id=id)
                current_pos = model.society.get_pos(id=id)
                travel = ds.point_to_point(
                    point_1=previous_pos,
                    point_2=current_pos
                )
                total_travel += travel
            self.add_travel_distance(value=total_travel)
            previous = deepcopy(model) # Push previous forward
            if report is True:
                print(f"Progress: {(i + 1) / n * 100:.1f}% complete")
        self.save()

    def save(self):
        """
        Save to file
        """
        data = self.serialize()
        file = JsonFile(path=self.result_directory, filename='measurement')
        file.save(data)
    
    def load(self):
        """
        Load to file
        """
        file = JsonFile(path=self.result_directory, filename='measurement')
        if file.exists() is False:
            raise ValueError("File not found:\n" + file.filepath)
        data = file.load()
        self.deserialize(data=data)

    def serialize(self) -> dict:
        """
        Serialize
        """
        return {
            'accessibility': self.accessibility.serialize(),
            'travel_distance': self.travel_distance.serialize(),
            'times': self.times,
            'type': self.type,
        }
    
    def deserialize(self, data: dict) -> None:
        """
        Deerialize
        """
        self.accessibility.deserialize(data['accessibility'])
        self.travel_distance.deserialize(data['travel_distance'])
        self.times = data['times']


if __name__ == "__main__":
    
    from piperabm.model.measurement import Measurement

    measure = Measurement()
    hour = 3600
    measure.add_time(0 * hour) # Base

    # 1
    measure.add_time(value=1*hour)
    measure.add_accessibility(id=1, value={'food': 1, 'water': 1, 'energy': 1})
    measure.add_accessibility(id=2, value={'food': 0.8, 'water': 0.7, 'energy': 0.6})
    measure.add_travel_distance(value=1.1)
    # 2
    measure.add_time(value=2*hour)
    measure.add_accessibility(id=1, value={'food': 0.9, 'water': 0.8, 'energy': 0.7})
    measure.add_accessibility(id=2, value={'food': 0.5, 'water': 0.6, 'energy': 0.4})
    measure.add_travel_distance(value=0.9)
    # 3
    measure.add_time(value=3*hour)
    measure.add_accessibility(id=1, value={'food': 0.8, 'water': 0.7, 'energy': 0.6})
    measure.add_accessibility(id=2, value={'food': 0.2, 'water': 0.4, 'energy': 0.3})
    measure.add_travel_distance(value=0.3)
    # 4
    measure.add_time(value=4*hour)
    measure.add_accessibility(id=1, value={'food': 0.7, 'water': 0.6, 'energy': 0.5})
    measure.add_accessibility(id=2, value={'food': 0, 'water': 0.3, 'energy': 0.2})
    measure.add_travel_distance(value=0.46)
    # 5
    measure.add_time(value=5*hour)
    measure.add_accessibility(id=1, value={'food': 0.6, 'water': 0.5, 'energy': 0.4})
    measure.add_accessibility(id=2, value={'food': 0, 'water': 0.3, 'energy': 0.2})
    measure.add_travel_distance(value=0.2)

    agents = 'all'
    resources = 'all'
    _from = None
    _to = None
    print("travel distances: ", measure.travel_distance(_from=_from, _to=_to))
    print("accessibilities: ", measure.accessibility(agents=agents, resources=resources, _from=_from, _to=_to))
    print("average: ", measure.accessibility.average(agents=agents, resources=resources))
    


from piperabm.society.actions.action.move.move import Move
from piperabm.society.actions.action.stay.stay import Stay


def action_deserialize(dictionary, queue):
    if dictionary["type"] == "move":
        object = Move()
    elif dictionary["type"] == "stay":
        object = Stay()
    object.deserialize(dictionary)
    object.queue = queue
    return object


if __name__ == "__main__":

    from piperabm.society.samples.society_1 import model

    agent_id = model.society.agents[0]
    destination_id = 2
    action_queue = model.society.actions[agent_id]
    current_node = model.society.get_current_node(agent_id)
    path = model.infrastructure.path(id_start=current_node, id_end=destination_id)
    move = Move(
        action_queue=action_queue,
        path=path,
        usage=1
    )
    action_queue.add(move)
    action_serialized = move.serialize()
    action_deserialized = action_deserialize(action_serialized, action_queue)
    
    print(action_deserialized.serialize() == move.serialize())


from piperabm.tools.print.stat import Print


class Stat(Print):
    """
    Stats of the network
    """

    @property
    def stat(self):
        """
        Return stats of the network
        """
        return {
            'node': {
                'junction': len(self.junctions),
                'home': len(self.homes),
                'market': len(self.markets),
            },
            'edge': {
                'street': len(self.streets),
                'neighborhood_access': len(self.neighborhood_accesses),
            },
        }
    

if __name__ == "__main__":

    from piperabm.infrastructure import Infrastructure

    infrastructure = Infrastructure()
    infrastructure.add_street(pos_1=[0, 0], pos_2=[10, 10])
    infrastructure.bake()
    print(infrastructure)


"""
10*10 grid of equally spaced points in xy plane are projected
"""

import numpy as np
import matplotlib.pyplot as plt

from piperabm.tools.coordinate.projection.latlong_xy import xy_latlong


latitude_0 = 70
longitude_0 = -150
DELTA = 1000000 # m
num = 10

# Calculate x and y values
xs_linspace = list(np.linspace(0 - DELTA/2, 0 + DELTA/2, num))
ys_linspace = list(np.linspace(0 - DELTA/2, 0 + DELTA/2, num))
xs_mesh, ys_mesh = np.meshgrid(xs_linspace, ys_linspace)
xs = xs_mesh.flatten()
ys = ys_mesh.flatten()

# Calculate latitude and longitude values
latitudes = []
longitudes = []
for i in range(num ** 2):
    latitude, longitude = xy_latlong(latitude_0, longitude_0, xs[i], ys[i])
    latitudes.append(latitude)
    longitudes.append(longitude)

# Draw
fig, (ax1, ax2) = plt.subplots(1, 2)
fig.suptitle('xy_to_latlong')

ax1.scatter(xs, ys)
ax1.set_title('before')
ax1.set_xlabel('x (meters)')
ax1.set_ylabel('y (meters)')

ax2.scatter(longitudes, latitudes)
ax2.set_title('after')
ax2.set_xlabel('longitude (degree)')
ax2.set_ylabel('latitude (degree)')

# Display the plots
plt.tight_layout()
plt.show()


from scipy.interpolate import griddata
from scipy.interpolate import LinearNDInterpolator
from scipy.interpolate import NearestNDInterpolator
from scipy.interpolate import RBFInterpolator


class Fit3D:
    """
    Interpolate 3d data
    """

    def __init__(self):
        self.xs = []
        self.ys = []
        self.zs = []

    def add(self, x: float, y: float, z: float):
        """
        Add new data point
        """
        self.xs.append(x)
        self.ys.append(y)
        self.zs.append(z)

    @property
    def len(self):
        """
        Number of data points
        """
        return len(self.zs)
    
    @property
    def points(self):
        """
        Prepare data for interpolation
        """
        return list(zip(self.xs, self.ys))

    def z(
            self,
            x: float,
            y: float,
            method: str = 'RBFInterpolator'
        ):
        """
        Return interpolated value
        """
        if self.len == 0:
            result = 0
        else:
            if method == 'griddata':
                result = griddata(
                    points=(self.xs, self.ys),
                    values=self.zs,
                    xi=(x, y),
                    method='cubic'
                )
            elif method == 'LinearNDInterpolator':
                interpolator = LinearNDInterpolator(self.points, self.zs)
                result = interpolator(x, y)
            elif method == 'NearestNDInterpolator':
                interpolator = NearestNDInterpolator(self.points, self.zs)
                result = interpolator(x, y)
            elif method == 'RBFInterpolator':
                rbf = RBFInterpolator(self.points, self.zs, kernel='cubic')
                return rbf([(x, y)])[0]
        return result
    
    def serialize(self):
        """
        Serialize
        """
        return {
            'xs': self.xs,
            'ys': self.ys,
            'zs': self.zs
        }
    
    def deserialize(self, dictionary: dict):
        """
        Deserialize
        """
        self.xs = dictionary['xs']
        self.ys = dictionary['ys']
        self.zs = dictionary['zs']


if __name__ == "__main__":
    interpolate = Fit3D()
    interpolate.add(x=10, y=0, z=10)
    interpolate.add(x=-10, y=0, z=-10)
    interpolate.add(x=0, y=10, z=5)
    interpolate.add(x=0, y=-10, z=-5)
    print("griddata: ", interpolate.z(x=0, y=0, method='griddata'))
    print("LinearNDInterpolator: ", interpolate.z(x=0, y=0, method='LinearNDInterpolator'))
    print("NearestNDInterpolator: ", interpolate.z(x=0, y=0, method='NearestNDInterpolator'))
    print("RBFInterpolator: ", interpolate.z(x=0, y=0, method='RBFInterpolator'))


class NxSet:
    """
    General set methods for networkx
    """

    def set_node_attribute(self, id: int, attribute: str, value=None) -> None:
        """
        Set node attribute in networkx graph
        """
        self.G.nodes[id][attribute] = value

    def set_edge_attribute(self, ids: list, attribute: str, value=None) -> None:
        """
        Set edge attribute in networkx graph
        """
        self.G[ids[0]][ids[1]][attribute] = value


import networkx as nx
import matplotlib.pyplot as plt

from piperabm.infrastructure.graphics.style import infrastructure_style


class Graphics:
    """
    Handle graphics
    """

    def fig(
            self,
            clf: bool = False
        ):
        """
        Add infrastructure elements to plt fig ax
        """
        if clf is True:
            plt.clf()
        ax = plt.gca()
        ax.set_aspect("equal")

        # Nodes
        pos_dict = {}
        node_color_list = []
        node_size_list = []
        node_label_dict = {}
        nodes = self.nodes
        for node_id in nodes:
            # Position
            pos_dict[node_id] = self.get_pos(id=node_id)
            # Color
            color = infrastructure_style['node'][self.get_node_type(node_id)]['color']
            node_color_list.append(color)
            # Size
            size = infrastructure_style['node'][self.get_node_type(node_id)]['radius']
            node_size_list.append(size)
            # Label
            node_label_dict[node_id] = self.get_node_name(id=node_id)

        # Edges
        edge_color_list = []
        edges = self.edges
        for edge_ids in edges:
            # Color
            color = infrastructure_style['edge'][self.get_edge_type(ids=edge_ids)]['color']
            edge_color_list.append(color)

        # Draw
        nx.draw_networkx(
            self.G,
            nodelist=nodes,
            pos=pos_dict,
            node_color=node_color_list,
            node_size=node_size_list,
            labels=node_label_dict,
            font_size=infrastructure_style['font'],
            edgelist=edges,
            edge_color=edge_color_list,
            ax=ax
        )
        return plt.gcf()

    def show(self):
        """
        Show infrastructure elements
        """
        fig = self.fig()
        plt.show()


if __name__ == "__main__":

    from piperabm.infrastructure.samples.infrastructure_2 import model

    model.infrastructure.show()


import numpy as np


def magnitude(vector) -> float:
    """
    Retrun magnitude of a vector
    """
    if vector is None:
        raise ValueError
    return float(np.sqrt(np.sum(np.square(vector))))


if __name__ == "__main__":
    vector = [3, 4]
    print(magnitude(vector))
    


from piperabm.tools.coordinate import distance as ds


class Rule0:
    """
    Condition for node to node proximity
    """

    name = "rule 0"

    def __init__(self, infrastructure, proximity_radius: float = 1):
        self.infrastructure = infrastructure
        self.proximity_radius = proximity_radius

    def check(self, node_id, other_node_id):
        result = False
        edge_constraint = True
        if self.infrastructure.has_edge(ids=[node_id, other_node_id]):
            if self.infrastructure.get_edge_type(ids=[node_id, other_node_id]) == "neighborhood_access":
                edge_constraint = False
        if edge_constraint is True:
            distance = ds.point_to_point(
                point_1=self.infrastructure.get_pos(node_id),
                point_2=self.infrastructure.get_pos(other_node_id)
            )
            if distance < self.proximity_radius:
                result = True
        return result
    
    def apply(self, node_id, other_node_id, report=False):
        # Remove any edge object between nodes
        if self.infrastructure.has_edge(ids=[node_id, other_node_id]) is True:
            self.infrastructure.remove_edge(ids=[node_id, other_node_id], report=report)

        # Merge nodes
        pos_node = self.infrastructure.get_pos(node_id)
        pos_other_node = self.infrastructure.get_pos(other_node_id)
        pos_new = [
            (pos_node[0] + pos_other_node[0]) / 2,
            (pos_node[1] + pos_other_node[1]) / 2,
        ]
        new_id = self.infrastructure.add_junction(pos=pos_new)
        self.infrastructure.replace_node(node_id, new_id, report=report)
        self.infrastructure.replace_node(other_node_id, new_id, report=report)
    
    def find(self, report=False):
        anything_happened = False
        nodes = self.infrastructure.junctions
        for node_id in nodes:
            for other_node_id in nodes:
                if node_id != other_node_id:
                    if self.check(node_id, other_node_id) is True:
                        if report is True:
                            print(f"\n# {self.name} for nodes at {self.infrastructure.get_pos(node_id)} and {self.infrastructure.get_pos(other_node_id)}")
                        self.apply(node_id, other_node_id, report=report)
                        # Inform an activity
                        anything_happened = True
                # Inform an activity
                if anything_happened is True:
                    break
            # Inform an activity     
            if anything_happened is True:
                break
        return anything_happened
    

if __name__ == "__main__":

    from piperabm.infrastructure import Infrastructure

    infrastructure = Infrastructure()
    infrastructure.add_street(pos_1=[0, 0.1], pos_2=[0, 10])
    infrastructure.add_street(pos_1=[0.1, 0], pos_2=[10, 0])
    
    rule = Rule0(infrastructure, proximity_radius=1)
    rule.find(report=True)
    print(infrastructure)
    


import numpy as np

from piperabm.tools.vector import vector as vc


def point_to_line(
        point,
        line_point_1,
        line_point_2,
        vector: bool = False,
        ndarray: bool = False
    ):
    """
    Finds the vector connecting a point to a line that has the smallest distance.
    
    :param point: The point (np.ndarray).
    :param line_point_1: First point on the line (np.ndarray).
    :param line_point_2: Second point on the line (np.ndarray).
    :return: The vector from the point to the closest point on the line.
    """
    result = None

    point = np.array(point)
    line_point_1 = np.array(line_point_1)
    line_point_2 = np.array(line_point_2)

    # Line vector
    line_vec = line_point_2 - line_point_1

    # Vector from line_point_1 to the point
    point_vec = point - line_point_1

    # Project point_vec onto line_vec
    projection = np.dot(point_vec, line_vec) / np.dot(line_vec, line_vec) * line_vec

    # The perpendicular vector from the point to the line
    result = projection - point_vec

    if vector is True:
        if ndarray is True:
            pass
        else:
            result = list(result)
    else:
        result = vc.magnitude(result)

    return result


if __name__ == '__main__':
    point = [3, 4]
    line_point_1 = [0, 0]
    line_point_2 = [2, 0]
    vector = point_to_line(point, line_point_1, line_point_2)
    print(vector)


from piperabm.model.graphics.style import *


node_radius = DEFAULT_NODE_RADIUS * 2
font_size = DEFAULT_FONT_SIZE + 2

#FONT_SIZE = 8
#NODE_ITEM_DEFAULT_RADIUS = 5


society_node_style = {
    "agent": {
        "color": {
            "dead": "r",
            "alive": "magenta",
        },
        #"shape": "x",
        "size": node_radius,
    }
}

society_edge_style = {
    "family": {
        "color": "b",
    },
    "neighbor": {
        "color": "b",
    },
    "friend": {
        "color": "b",
    },
}

society_style = {
    "font": font_size,
    "node": society_node_style,
    "edge": society_edge_style,
}


if __name__ == "__main__":
    print(society_style)


import numpy as np


class Mercator:   

    @staticmethod
    def project(latitude_degree=0, longitude_degree=0, radius=1):
        """
        Project using Mercator formula to cartesian coordinates.
        """
        latitude_rad = np.radians(latitude_degree)
        longitude_rad = np.radians(longitude_degree)
        x = radius * longitude_rad
        y = radius * np.log(np.tan(np.pi / 4 + latitude_rad / 2))
        return x, y
    
    @staticmethod
    def inverse(x=0, y=0, radius=1):
        """
        Inverse project from Cartesian coordinates to (latitude, longitude) in degrees.
        """
        longitude_rad = x / radius
        latitude_rad = 2 * np.arctan(np.exp(y / radius)) - np.pi / 2
        latitude_degree = np.degrees(latitude_rad)
        longitude_degree = np.degrees(longitude_rad)
        return latitude_degree, longitude_degree
    

if __name__ == "__main__":
    latitude = 70
    longitude = -150
    radius = 6378
    x, y = Mercator.project(latitude, longitude, radius)
    print(x, y)
    latitude, longitude = Mercator.inverse(x, y, radius)
    print(latitude, longitude)




import os

import piperabm as pa
from piperabm.society.samples.society_2 import model


# Setup
path = os.path.dirname(os.path.realpath(__file__))
model.path = path

# Run
print(">>> Running...")
model.run(n=48, save=True, resume=False, report=True, step_size=4*3600)

# Measure
print(">>> Measuring...")
measurement = pa.Measurement(path=path)
measurement.measure(resume=False, report=True)


class Update:
    """
    Update the network
    """

    def update(self, duration: float):
        """
        Update the network
        """
        # Update degradation from climate change (streets only)
        rate = 0.00001
        for ids in self.streets:
            # Update weather impact
            climate_impact = self.get_climate_impact(ids=ids)
            climate_impact += rate * duration
            self.set_climate_impact(ids=ids, value=climate_impact)
            # Update corresponding edge
            self.update_adjusted_length(ids=ids)


if __name__ == "__main__":

    from piperabm.society.samples.society_1 import model

    model.update(10)
    model.show()


from piperabm.society.actions.action.move.track import Track
from piperabm.tools.print.serialized import Print


class Move(Print):
    """
    Move action
    """

    type = "move"

    def __init__(
            self,
            action_queue=None,
            path: list = [],
            usage: float = 1,
        ):
        super().__init__()
        self.action_queue = action_queue  # Binding
        self.tracks = self.path_to_tracks(path)
        self.usage = usage

    @property
    def society(self):
        """
        Alias to access society
        """
        return self.action_queue.society
    
    @property
    def model(self):
        """
        Alias to access model
        """
        return self.society.model
    
    @property
    def infrastructure(self):
        """
        Alias to access infrastructure
        """
        return self.model.infrastructure
    
    @property
    def agent_id(self):
        """
        Alias to access agent id
        """
        return self.action_queue.agent_id

    def path_to_tracks(self, path):
        """
        Convert path to track segments
        """
        tracks = []
        for i, _ in enumerate(path):
            if i != 0:
                id_start = path[i-1]
                id_end = path[i]
                track = Track(
                    action=self, # Binding
                    id_start=id_start,
                    id_end=id_end
                )
                track.preprocess()
                tracks.append(track)
        return tracks
    
    @property
    def total_duration(self):
        """
        Return how long the action will take
        """
        total = 0
        for track in self.tracks:
            total += track.total_duration
        return total
    
    @property
    def destination(self):
        """
        Destionation id
        """
        last_track = self.tracks[-1]
        return last_track.id_end

    @property
    def active_track(self):
        """
        Return the track which is currently in progress
        """
        result = None
        for track in self.tracks:
            #print(track.done)
            if track.done is False:
                result = track
                break
        return result

    @property
    def done(self):
        """
        Check whether the action is already complete
        """
        result = True
        for track in self.tracks:
            status = track.done
            if status is False:
                result = False
                break
        return result

    @property
    def elapsed(self):
        """
        Return how long has been passed
        """
        result = 0
        for track in self.tracks:
            result += track.elapsed
        return result

    @property
    def remaining(self):
        """
        Return how long is still remaining
        """
        result = 0
        for track in self.tracks:
            result += track.remaining
        return result

    def reverse(self):
        """
        Create a reversed move action
        """
        reversed_tracks = []
        for track in self.tracks:
            reversed_tracks.append(track.reverse())
        reversed_tracks = list(reversed(reversed_tracks))
        reversed_move = Move(
            action_queue=self.action_queue,
            path=[],
            usage=self.usage
        )
        reversed_move.tracks = reversed_tracks
        return reversed_move

    def update(self, duration, measure: bool = False):
        """
        Update status of action
        """
        self.society.set_current_node(self.agent_id, None)
        excess_delta_time = duration
        while excess_delta_time > 0:
            track = self.active_track
            if track is not None:
                excess_delta_time = track.update(excess_delta_time, measure=measure)
            else: # Action ended
                self.society.set_current_node(self.agent_id, self.destination)
                break
        return excess_delta_time

    def serialize(self) -> dict:
        """
        Serialize
        """
        data = {}
        data['type'] = self.type
        tracks_serialized = []
        for track in self.tracks:
            track_serialized = track.serialize()
            tracks_serialized.append(track_serialized)
        data['tracks'] = tracks_serialized
        data['usage'] = self.usage
        return data

    def deserialize(self, data: dict) -> None:
        """
        Deserialize
        """
        tracks_serialized = data['tracks']
        for track_serialized in tracks_serialized:
            track = Track(
                action=self,
                id_start=track_serialized['id_start'],
                id_end=track_serialized['id_end']
            )
            track.preprocess(track_serialized)
            self.tracks.append(track)
        self.usage = data['usage']


if __name__ == "__main__":

    from piperabm.society.samples.society_1 import model

    agent_id = model.society.agents[0]
    destination_id = 2
    action_queue = model.society.actions[agent_id]
    path = model.infrastructure.path(
        id_start=model.society.get_current_node(id=agent_id),
        id_end=destination_id
    )
    move = Move(
        action_queue=action_queue,
        path=path,
        usage=1
    )
    action_queue.add(move)

    #print(move)

    print(f"total duration: {move.total_duration}")
    street = model.infrastructure.streets[0]
    print(f"usage impact: {model.infrastructure.get_usage_impact(ids=street)}")
    print(f"time: {model.time}, pos: {model.society.get_pos(agent_id)}")
    
    model.update(duration=28)
    
    print(f"time: {model.time}, pos: {model.society.get_pos(agent_id)}")
    
    model.update(duration=28)
    
    print(f"time: {model.time}, pos: {model.society.get_pos(agent_id)}")
    print(f"usage impact: {model.infrastructure.get_usage_impact(ids=street)}")



import matplotlib.pyplot as plt

from piperabm.model.graphics.animation import Animation


class Graphics:
    """
    Handle graphics
    """

    def fig(
            self,
            relationships=False,
            clf=True
        ):
        """
        Add model elements to plt fig ax
        """
        if clf is True:
            plt.clf()
        ax = plt.gca()
        ax.set_aspect("equal")

        # Draw infrastructure
        fig = self.infrastructure.fig(
            clf=False
        )

        # Draw society
        fig = self.society.fig(
            relationships=relationships,
            clf=False
        )

        return fig

    def show(self, relationships=False):
        """
        Show model elements
        """
        fig = self.fig(
            relationships=relationships
        )
        plt.show()

    def animate(self):
        animation = Animation(path=self.result_directory)
        self.load_initial()

        # First frame
        fig = self.fig()
        animation.add_figure(fig)

        # Apply deltas and create frames
        deltas = self.load_deltas()
        for delta in deltas:
            self.apply_delta(delta)
            fig = self.fig()
            animation.add_figure(fig)
    
        # Render
        animation.render(output_file='animation')


if __name__ == "__main__":

    from piperabm.society.samples import model_2 as model

    model.show(relationships=['neighbor'])


from piperabm.society.actions.action import Move
from piperabm.society.actions.action.stay.stay import Stay
from piperabm.tools.print.serialized import Print


class ActionQueue(Print):
    """
    Action queue to manage the to-do list
    """

    type = "queue"

    def __init__(
            self,
            agent_id: int = None
        ):
        super().__init__()
        self.society = None  # Binding
        self.library = []
        self.agent_id = agent_id
        
    def add(self, *actions):
        """
        Add new action(s) to the queue
        """
        for element in actions:
            if isinstance(element, list):
                for action in element:
                    self.add(action)
            else:
                action = element
                action.action_queue = self  # Binding
                self.library.append(action)

    def reset(self):
        """
        Reset the queue
        """
        self.library = []

    @property
    def undones(self):
        """
        Find undone actions from end
        """
        undone_actions = []
        for action in reversed(self.library):
            if action.done:
                break  # Stop the iteration if an action is done, as per the list"s structure.
            undone_actions.append(action)
        return list(reversed(undone_actions))
    
    @property
    def done(self):
        """
        Whether all actions are done or not
        """
        result = None
        if len(self.undones) > 0:
            result = False
        else:
            result = True
        return result
    
    @property
    def remaining(self):
        """
        Estimate the remaining time to complete undone tasks
        """
        total = 0
        undone_actions = self.undones
        for action in undone_actions:
            remaining = action.remaining
            total += remaining
        return total
    
    @property
    def elapsed(self):
        """
        Estimate the remaining time to complete undone tasks
        """
        total = 0
        for action in self.library:
            remaining = action.elapsed
            total += remaining
        return total
    
    @property
    def total_duration(self):
        """
        Return how long all the actions will take
        """
        total = 0
        for action in self.library:
            duration = action.total_duration
            total += duration
        return total

    def update(self, duration, measure: bool = False):
        """
        Update the queue
        """
        for action in self.undones:
            duration = action.update(duration, measure=measure)
            if duration == 0:
                break

    def serialize(self):
        """
        Serialize
        """
        data = {}
        library_serialized = []
        for action in self.library:
            action_serialized = action.serialize()
            library_serialized.append(action_serialized)
        data["library"] = library_serialized
        data["agent_id"] = self.agent_id
        data["type"] = self.type
        return data
    
    def deserialize(self, data: dict) -> None:
        """
        Deserialize
        """
        if data["type"] != self.type:
            raise ValueError
        library_serialized = data["library"]
        for action_serialized in library_serialized:
            if "type" in action_serialized:
                if action_serialized["type"] == "move":
                    action = Move(action_queue=self)
                elif action_serialized["type"] == "stay":
                    action = Stay(action_queue=self)
                action.deserialize(action_serialized)
                self.library.append(action)
    

if __name__ == "__main__":

    from piperabm.society.samples.society_1 import model

    agent_id = model.society.agents[0]
    destination_id = 2
    action_queue = model.society.actions[agent_id]
    path = model.infrastructure.path(
        id_start=model.society.get_current_node(id=agent_id),
        id_end=destination_id
    )
    move = Move(
        action_queue=action_queue,
        path=path,
        usage=1
    )

    agent_id = model.society.agents[0]
    destination_id = 2
    action_queue = model.society.actions[agent_id]
    model.society.go_and_comeback_and_stay(
        action_queue=action_queue,
        move_go=move,
        stay_length=100,
    )
    
    #print(model.society.actions[agent_id])
    print(f"time: {model.time}, pos: {model.society.get_pos(agent_id)}")

    model.update(duration=30)

    print(f"time: {model.time}, pos: {model.society.get_pos(agent_id)}")

    model.update(duration=30)

    print(f"time: {model.time}, pos: {model.society.get_pos(agent_id)}")

    model.update(duration=120)

    print(f"time: {model.time}, pos: {model.society.get_pos(agent_id)}")

    model.update(duration=60)

    print(f"time: {model.time}, pos: {model.society.get_pos(agent_id)}")
    


import numpy as np

from piperabm.tools.coordinate import distance as ds
from piperabm.tools.vector import vector as vc
from piperabm.infrastructure.grammar.rules.rule_1 import Rule1


class Rule3:
    """
    Condition for connecting isolated non-junction items to the rest
    """

    name = "rule 3"

    def __init__(self, infrastructure, proximity_radius: float = 1, search_radius: float = None):
        self.infrastructure = infrastructure
        self.search_radius = search_radius
        self.proximity_radius = proximity_radius

    def check(self, node_id):
        result = False
        # Has to be isolated
        if self.infrastructure.is_isolate(node_id) and \
            len(self.infrastructure.edges) > 0:
            result = True
        return result
    
    def apply(self, node_id, report: bool = False):
        node_pos = self.infrastructure.get_pos(node_id)
        edges_ids = self.infrastructure.edges
        distances_info = []
        # Filter edges
        if self.search_radius is None:
            edges_ids_nearby = edges_ids
        else:
            edges_ids_nearby = self.infrastructure.edges_closer_than(
                pos=node_pos,
                max_distance=self.search_radius,
                edges_ids=edges_ids,
            )
        # Find the best location to target in the filtered edges
        for edge_ids in edges_ids_nearby:
            # Only perpendicular distance from edge
            distance_vector_edge = ds.point_to_line(
                point=node_pos,
                line=[
                    self.infrastructure.get_pos(edge_ids[0]),
                    self.infrastructure.get_pos(edge_ids[1])
                ],
                segment=True,
                vector=True,
                perpendicular_only=True,
            )
            edge_info = None
            if distance_vector_edge is not None:
                edge_info = {
                    'id': edge_ids,
                    'vector': distance_vector_edge,
                    'distance': vc.magnitude(distance_vector_edge),
                    'type': 'edge',
                }
            id_1 = edge_ids[0]
            distance_vector_node_1 = ds.point_to_point(
                point_1=node_pos,
                point_2=self.infrastructure.get_pos(id_1),
                vector=True
            )
            node_info_1 = {
                'id': id_1,
                'vector': distance_vector_node_1,
                'distance': vc.magnitude(distance_vector_node_1),
                'type': 'node',
            }
            id_2 = edge_ids[1]
            distance_vector_node_2 = ds.point_to_point(
                point_1=node_pos,
                point_2=self.infrastructure.get_pos(id_2),
                vector=True
            )
            node_info_2 = {
                'id': id_2,
                'vector': distance_vector_node_2,
                'distance': vc.magnitude(distance_vector_node_2),
                'type': 'node',
            }
            if node_info_1['distance'] < node_info_2['distance']:
                possible_top_node_info = node_info_1
            else:
                possible_top_node_info = node_info_2
            if edge_info is None:
                possible_target_info = possible_top_node_info
            elif edge_info['distance'] < possible_top_node_info['distance']:
                possible_target_info = edge_info
            else:
                possible_target_info = possible_top_node_info
            distances_info.append(possible_target_info)

        # Top target info
        target_info = None
        for possible_target_info in distances_info:
            if target_info is None or \
                possible_target_info['distance'] < target_info['distance']:
                target_info = possible_target_info

        # Node as target
        if target_info['type'] == 'node':
            self.infrastructure.add_neighborhood_access(
                node_id,
                target_info['id'],
                report=report
            )
        # Edge as target
        elif target_info['type'] == 'edge':
            pos_1 = self.infrastructure.get_pos(node_id)
            pos_2 = list(np.array(pos_1) + np.array(target_info['vector']))
            new_node_id = self.infrastructure.add_junction(pos=pos_2, report=report)
            self.infrastructure.add_neighborhood_access(
                node_id,
                new_node_id,
                report=report
            )
            rule_1 = Rule1(self.infrastructure, proximity_radius=self.proximity_radius)
            rule_1.apply(node_id=new_node_id, edge_ids=target_info['id'], report=report)
        self.infrastructure.baked_streets = True
    
    def find(self, report=False):
        anything_happened = False
        nodes = self.infrastructure.nonjunctions
        for node_id in nodes:
            if self.check(node_id) is True:
                if report is True:
                    print(f"# {self.name}:")
                self.apply(node_id, report)
                # Inform an activity
                anything_happened = True
            # Inform an activity
            if anything_happened is True:
                break
        return anything_happened
    

if __name__ == "__main__":

    from piperabm.infrastructure import Infrastructure

    infrastructure = Infrastructure()
    infrastructure.add_street(pos_1=[0, 0], pos_2=[10, 0])
    infrastructure.add_home(pos=[5, 4], id=1)

    rule = Rule3(infrastructure, proximity_radius=1, search_radius=None)
    rule.find(report=True)
    print(infrastructure)


import os

from piperabm.model.measurement import Measurement


path = os.path.dirname(os.path.realpath(__file__))

measure = Measurement(path=path, name='model')
measure.load()

agents = 'all'
resources = 'all'
_from = None
_to = None
print("travel distances: ", measure.travel_distance(_from=_from, _to=_to))
print("accessibilities: ", measure.accessibility(agents=agents, resources=resources, _from=_from, _to=_to))
print("average: ", measure.accessibility.average(agents=agents, resources=resources))


import networkx as nx


class Path:
    """
    Path finding algorythm
    """

    def has_path(self, id_start: int, id_end: int) -> bool:
        """
        Rapidly check if there is any path
        """
        return nx.has_path(
            self.G,
            source=id_start,
            target=id_end
        )

    def path(self, id_start: int, id_end: int) -> list:
        """
        Path finding algorythm using A_star
        """
        return nx.astar_path(
            self.G,
            source=id_start,
            target=id_end,
            heuristic=self.heuristic_paths.estimated_distance,
            weight="adjusted_length"
        )


if __name__ == "__main__":

    from piperabm.infrastructure.samples.infrastructure_1 import model

    has_path = model.infrastructure.has_path(id_start=1, id_end=2)
    print("has path? :", has_path)
    
    path = model.infrastructure.path(id_start=1, id_end=2)
    print("path: ", path)


import networkx as nx
import numpy as np

from piperabm.tools.coordinate import distance as ds
from piperabm.infrastructure.query.add import Add
from piperabm.infrastructure.query.get import Get
from piperabm.infrastructure.query.set import Set


class Query(Add, Get, Set):
    """
    Query network elements
    """

    def has_node(self, id: int) -> bool:
        """
        Check whether the network already contains the node
        """
        return self.G.has_node(id)

    def has_edge(self, ids: list) -> bool:
        """
        Check whether the network already contains the edge
        """
        return self.G.has_edge(*ids)

    def is_isolate(self, id: int) -> bool:
        """
        Check if the node is isolated
        """
        return nx.is_isolate(self.G, id)
    
    '''
    def filter_nodes_closer_than(self, id: int, distance: float, nodes: list = None) -> list:
        """
        Filter *nodes* that are within the *distance* from *id*
        """
        result = []
        if nodes is None:
            nodes = self.nodes
        for node_id in nodes:
            if distance >= self.heuristic_paths.estimated_distance(id_start=id, id_end=node_id):
                result.append(node_id)
        return result
    '''
        
    def replace_node(self, id: int, new_id: int, report: int = False) -> None:
        """
        Replace a node with another node
        """
        # Find all adjacent edges
        edges_ids = self.edges_from(id)
        # Apply change to adjacent edges
        for edge_ids in edges_ids:
            # Create new edge
            if edge_ids[0] == id:
                new_edge_ids = [new_id, edge_ids[1]]
            else:
                new_edge_ids = [edge_ids[0], new_id]
            data = self.get_edge_attributes(ids=edge_ids)
            data['length'] = ds.point_to_point(
                    self.get_pos(new_edge_ids[0]),
                    self.get_pos(new_edge_ids[1])
                )
            adjustment_factor = self.calculate_adjustment_factor(
                usage_impact=data['usage_impact'],
                climate_impact=data['climate_impact']
            )
            data['adjusted_length'] = self.calculate_adjusted_length(
                length=data['length'],
                adjustment_factor=adjustment_factor,
            )
            self.G.add_edge(
                new_edge_ids[0],
                new_edge_ids[1],
                **data
            )
            if report is True:
                print(f">>> {type} edge at positions {self.get_pos(new_edge_ids[0])} - {self.get_pos(new_edge_ids[1])} added.")
            # Remove old edge
            self.remove_edge(ids=edge_ids, report=report)
        # Remove old node
        self.remove_node(id, report=report)

    def impact(self, edges: list = []):
        """
        Impact the network by removing a list of edges
        """
        if self.baked is False:
            print("First bake the model")
            raise ValueError
        for ids in edges:
            self.remove_edge(ids=ids)
            id_1 = ids[0]
            id_2 = ids[1]
            if self.get_node_type(id=id_1) == 'junction' and \
                self.is_isolate(id=id_1):
                self.remove_node(id=id_1)
            if self.get_node_type(id=id_2) == 'junction' and \
                self.is_isolate(id=id_2):
                self.remove_node(id=id_2)

    def random_edges(self, percent: float = 0):
        """
        Filter random edges by their length percentage
        """
        if percent > 100:
            raise ValueError("enter a value between 0 and 100")
        edges_ids = self.streets
        total_length = 0
        edges_info = []
        for edge_ids in edges_ids:
            length = self.get_edge_attribute(ids=edge_ids, attribute='length')
            total_length += length
            edge_info = {
                'ids': edge_ids,
                'length': length
            }
            edges_info.append(edge_info)
        remaining_length = (percent / 100) * total_length
        result = []
        np.random.shuffle(edges_info)
        for edge_info in edges_info:
            remaining_length -= edge_info['length']
            if remaining_length < 0:
                break
            else:
                result.append(edge_info['ids'])
        return result
    
    def remove_edge(self, ids: list = None, report: bool = False):
        """
        Remove edge
        """
        if report is True:
            print(f">>> {self.get_edge_type(ids=ids)} edge at {self.get_pos(ids[0])} - {self.get_pos(ids[1])} removed.")
        self.G.remove_edge(*ids)

    def remove_node(self, id: int, report: bool = False):
        """
        Remove node
        """
        if report is True:
            print(f">>> {self.get_node_type(id=id)} node at position {self.get_pos(id)} removed.")
        self.G.remove_node(id)
    
    def nodes_closer_than(self, id: int, search_radius: float = 0, nodes: list = None, include_self: bool = False):
        """
        Filter *nodes* that are within the *distance* from *id*
        """
        if nodes is None:
            nodes = self.nodes
        result = []
        for node_id in nodes:
            if search_radius >= self.heuristic_paths.estimated_distance(id_start=id, id_end=node_id):
                if include_self is False:
                    if node_id != id:
                        result.append(node_id)
                else:
                    result.append(node_id)
        return result
    
    @property
    def junctions(self) -> list:
        """
        Return all junction nodes
        """
        try:
            return [n for n, attr in self.G.nodes(data=True) if attr.get('type') == 'junction']
        except:
            return []
        
    @property
    def nonjunctions(self) -> list:
        """
        Return all nonjunction nodes
        """
        try:
            return [n for n, attr in self.G.nodes(data=True) if attr.get('type') != 'junction']
        except:
            return []
        
    @property
    def homes(self) -> list:
        """
        Return all homes nodes
        """
        try:
            return [n for n, attr in self.G.nodes(data=True) if attr.get('type') == 'home']
        except:
            return []

    @property
    def markets(self) -> list:
        """
        Return all market nodes
        """
        try:
            return [n for n, attr in self.G.nodes(data=True) if attr.get('type') == 'market']
        except:
            return []
    
    @property
    def streets(self) -> list:
        """
        Return all street edges
        """
        try:
            return [(u, v) for u, v, attr in self.G.edges(data=True) if attr.get('type') == 'street']
        except:
            return []
    
    @property
    def neighborhood_accesses(self) -> list:
        """
        Return all neighborhood access edges
        """
        try:
            return [(u, v) for u, v, attr in self.G.edges(data=True) if attr.get('type') == 'neighborhood_access']
        except:
            return []


import json
import os


class JsonFile:
    """
    Json file handling
    """

    def __init__(self, path, filename: str, format: str = "json"):
        if path is None:
            print("path is not defined.")
            raise ValueError
        self.file = JsonFile._create_file_str(filename, format)
        self.file_temp = JsonFile._create_file_str(filename + "_" + "temp", format)
        self.filepath = os.path.join(path, self.file)
        self.filepath_temp = os.path.join(path, self.file_temp)

    def _create_file_str(filename: str, format: str):
        """
        Create a string for file name by attaching format
        """
        return filename + "." + format
    
    def save(self, data):
        """
        Save the data to file as json using atomic file writing
        """
        # Write data to the temporary file, overwriting if it already exists
        with open(self.filepath_temp, "w") as f:
            json.dump(data, f)

        # Remove the main file if it exists to prevent errors on renaming
        if self.exists():
            self.remove()

        # Rename the temporary file to the main file"s name
        os.rename(self.filepath_temp, self.filepath)

    def load(self):
        """
        Load the data from a file as json
        """
        try:
            with open(self.filepath, "r") as f:
                data = json.load(f)
        except FileNotFoundError:
            data = None
            print(f"The file {self.file} was not found.")
        data = JsonFile.convert_keys(data)
        return data
    
    def convert_keys(obj):
        """
        Recursively converts string keys to integers if possible in any JSON-like structure 
        which may include dictionaries and lists.
        """
        if isinstance(obj, dict):
            new_obj = {}
            for k, v in obj.items():
                new_key = int(k) if k.isdigit() else k  # Convert key if it's a digit
                new_obj[new_key] = JsonFile.convert_keys(v)  # Recurse into values
            return new_obj
        elif isinstance(obj, list):
            return [JsonFile.convert_keys(item) for item in obj]  # Apply recursively to each item in the list
        else:
            return obj  # Return the item itself if it's neither a dict nor a list
    
    def exists(self):
        """
        Check if the file already exists
        """
        return os.path.exists(self.filepath)

    def append(self, entry):
        """
        Add new entry to save file
        """
        data = self.load()
        if isinstance(data, list):
            data.append(entry)
        else:
            print("Data is not list.")
            raise ValueError
        self.save(data)

    def remove(self):
        """
        Remove file if exists
        """
        if self.exists():
            os.remove(self.filepath)


if __name__ == "__main__":
    path = os.path.dirname(os.path.realpath(__file__))
    filename = "sample"
    file = JsonFile(path, filename)

    data = []
    file.save(data)

    entry = {"a": 1}
    file.append(entry)

    data = file.load()
    print("Test: ", data == [{"a": 1}])

    file.remove()
    


import pprint


class Print:

    def __str__(self) -> str:
        """
        To print serialized format of object
        """
        data = self.serialize()
        txt = pprint.pformat(
            data,
            depth=5,
            compact=True,
            width=100,
        )
        return txt